// Copyright (C) 2005 Dave Griffiths
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#include <sys/time.h>
#include <iostream>
#include <fstream>
#include <cstdlib>
#include <unistd.h>
#include <cstdio>
#include <GL/glut.h>
#include <vector>
#include <list>
#include <png.h>

using namespace std;
static const int MAX_THREAD_LIFE=99999999;
static const unsigned int MAX_THREADS=9999;

typedef unsigned char data;

class Matrix
{
public:
	Matrix() 
	{ 
		m[0][0]=1; m[0][1]=0;
		m[1][0]=0; m[1][1]=1; 
	}
	
	Matrix &operator*=(Matrix &other);	
	void Transform(int &x, int &y);
	void Transform(data &x, data &y);
	void RotCW();
	void RotCCW();
	
	char m[2][2];
};

Matrix &Matrix::operator*=(Matrix &other)
{	
	Matrix t;
	for (int i=0; i<2; i++)
        for (int j=0; j<2; j++)
    		t.m[i][j]=m[0][j]*other.m[i][0]+m[1][j]*other.m[i][1];
			
	*this=t;
	
	return *this;
}

void Matrix::Transform(int &x, int &y)
{	
	int rx,ry;
	rx=x*m[0][0] + y*m[1][0];
	ry=x*m[0][1] + y*m[1][1];	
	x=rx; y=ry;
}

void Matrix::Transform(data &x, data &y)
{	
	data rx,ry;
	rx=x*m[0][0] + y*m[1][0];
	ry=x*m[0][1] + y*m[1][1];	
	x=rx; y=ry;
}

void Matrix::RotCW()
{
	 Matrix cw;
	 cw.m[0][0]=0; cw.m[0][1]=-1; 
	 cw.m[1][0]=1; cw.m[1][1]=0;
	 *this*=cw;
}

void Matrix::RotCCW()
{
	 Matrix cw;
	 cw.m[0][0]=0; cw.m[0][1]=1; 
	 cw.m[1][0]=-1; cw.m[1][1]=0;
	 *this*=cw;
}

//--------------------------------------------------------------------------------

class Quagmire
{
public:
	Quagmire(data *mem, unsigned int w, unsigned int h);
	
	enum Direction{NORTH,EAST,SOUTH,WEST};
	enum Rotation{CW,CCW};
	enum SpawnDesc{NONE,SPAWN_CW,SPAWN_CCW};

	class Thread
	{
	public:
		Thread(Quagmire *machine, unsigned int x, unsigned int y, Matrix space);
		void Process(SpawnDesc &desc);
		void RotCW();
		void RotCCW();
		void GetWorldPos(unsigned int &x, unsigned int&y);
		void IncPC();
	
		Matrix m_Space;		
		Direction m_Direction;
		bool m_DelMe;
			
	private:		
			
		data GetArg();
		data Peek(data x, data y);
		void Poke(data x, data y, data v);

		Quagmire *m_Machine;
		data m_WRegister;
		data m_PCX;
		data m_PCY;	
		unsigned int m_OriginX;
		unsigned int m_OriginY;	
		int m_Lifetime;
	};
	
	void Process();
	void Render(float zoom);
	data Peek(int x, int y);
	void Poke(int x, int y, data v);
	void SpawnThread(unsigned int x, unsigned int y, Matrix space);

	void RenderThreads(bool s) { m_RenderThreads=s; }
	
private:

	unsigned int m_Width,m_Height;
	data *m_Memory;
	data *m_View;
	list<Thread*> m_ThreadVec;
	bool m_RenderThreads;
};



Quagmire::Quagmire(data *mem, unsigned int w, unsigned int h) :
m_Width(w),
m_Height(h),
m_RenderThreads(true)
{
	m_Memory=mem;
	m_View=new data[w*h];
}

data Quagmire::Peek(int x, int y)
{	
	while (x<0) x+=m_Width;
	while (y<0) y+=m_Height;
	x=x%m_Width;
	y=y%m_Height;
	unsigned int index=x+y*m_Width;
	return m_Memory[index];
}

void Quagmire::Poke(int x, int y, data v)
{	
	while (x<0) x+=m_Width;
	while (y<0) y+=m_Height;
	x=x%m_Width;
	y=y%m_Height;	
	unsigned int index=x+y*m_Width;
	m_Memory[index]=v;
}

void Quagmire::SpawnThread(unsigned int x, unsigned int y, Matrix space)
{
	m_ThreadVec.push_back(new Thread(this,x,y,space));
}

void Quagmire::Process()
{
	SpawnDesc spawn;
	list<Thread*> newthrs;
	vector<list<Thread*>::iterator> delthr;

	for (list<Thread*>::iterator i=m_ThreadVec.begin(); i!=m_ThreadVec.end(); i++)
	{
		(*i)->Process(spawn);
		
		if (spawn!=NONE)
		{
			if (m_ThreadVec.size()<MAX_THREADS)
			{					
				Matrix childspace=(*i)->m_Space;			
				if (spawn==SPAWN_CW) childspace.RotCW();
				else childspace.RotCCW();

				unsigned int x,y;
				(*i)->GetWorldPos(x,y);

				Thread* newthr = new Thread(this, x, y, childspace);
				//newthr->IncPC(); // jump spawn instr
				newthrs.push_back(newthr);
			}		
			(*i)->IncPC();
		}
	
		if ((*i)->m_DelMe)
		{			
			delthr.push_back(i);
		}
	}
		
	for (vector<list<Thread*>::iterator>::iterator i=delthr.begin(); i!=delthr.end(); i++)
	{
		//if (m_ThreadVec.size()>0) 
		{
			delete **i;
			m_ThreadVec.erase(*i);
		}
	}
	
	
	for (list<Thread*>::iterator i=newthrs.begin(); i!=newthrs.end(); i++)
	{
		m_ThreadVec.push_back(*i);
	}
	
}


void Quagmire::Render(float zoom)
{	
	glPixelZoom(zoom,zoom);
			
	if (m_RenderThreads)
	{	
		memcpy(m_View,m_Memory,m_Width*m_Height*sizeof(data));
		
		for (list<Thread*>::iterator i=m_ThreadVec.begin(); i!=m_ThreadVec.end(); i++)
		{		
			unsigned int x,y;
			(*i)->GetWorldPos(x,y);
			m_View[(x%m_Width)+(y%m_Height)*m_Width]^=255;
		}
		
		glDrawPixels(m_Width,m_Height,GL_LUMINANCE,GL_UNSIGNED_BYTE,(void*)m_View);
	}
	else
	{
		glDrawPixels(m_Width,m_Height,GL_LUMINANCE,GL_UNSIGNED_BYTE,(void*)m_Memory);
	}
}

//---------------------------------------------------------------------------------------

Quagmire::Thread::Thread(Quagmire *machine, unsigned int x, unsigned int y, Matrix space): 
m_Space(space),
m_Direction(EAST),
m_DelMe(false),
m_Machine(machine), 
m_WRegister(0),
m_PCX(0),
m_PCY(0),
m_OriginX(x),
m_OriginY(y),
m_Lifetime(0)
{
	IncPC();
}

void Quagmire::Thread::RotCW()
{
	switch (m_Direction)
	{
		case NORTH: m_Direction=EAST;  break;
		case EAST:  m_Direction=SOUTH; break;
		case SOUTH: m_Direction=WEST;  break;
		case WEST:  m_Direction=NORTH; break;
	}
}

void Quagmire::Thread::RotCCW()
{
	switch (m_Direction)
	{
		case NORTH: m_Direction=WEST;  break;
		case EAST:  m_Direction=NORTH; break;
		case SOUTH: m_Direction=EAST;  break;
		case WEST:  m_Direction=SOUTH; break;
	}
}

void Quagmire::Thread::GetWorldPos(unsigned int &x, unsigned int &y)
{
	int xx=m_PCX;
	int yy=m_PCY;
	m_Space.Transform(xx,yy);
	x=xx+m_OriginX;
	y=yy+m_OriginY;
}

data Quagmire::Thread::GetArg()
{
	IncPC();
	unsigned int x=m_PCX;
	unsigned int y=m_PCY;
	GetWorldPos(x,y);
	return m_Machine->Peek(x,y);
}

data Quagmire::Thread::Peek(data x, data y)
{
	int xx=x,yy=y;
	m_Space.Transform(xx,yy);
	return m_Machine->Peek(xx+m_OriginX,yy+m_OriginY);
}

void Quagmire::Thread::Poke(data x, data y, data v)
{
	int xx=x,yy=y;
	m_Space.Transform(xx,yy);
	m_Machine->Poke(xx+m_OriginX,yy+m_OriginY,v);
}

void Quagmire::Thread::IncPC()
{
	switch (m_Direction)
	{
		case NORTH: m_PCY++; break;
		case EAST:  m_PCX++; break;
		case SOUTH: m_PCY--; break;
		case WEST:  m_PCX--; break;
	}
}


void Quagmire::Thread::Process(SpawnDesc &desc)
{
	data code=Peek(m_PCX,m_PCY);
	if (code!=255) code=code%0x18;
	desc=NONE;
	bool incpc=true;
	m_Lifetime++;
	if (m_Lifetime>MAX_THREAD_LIFE) m_DelMe=true;
	//cerr<<"code="<<(int)code<<" w="<<(int)m_WRegister<<endl;
	switch (code)
	{
		case 0x00: 
		{
			// nop
		} break;
		case 0x01: 		
		{
			// movlw (write literal to w)
			IncPC();
			m_WRegister=Peek(m_PCX,m_PCY);
		} break;
		case 0x02: 		
		{
			// movf x y (w to mem location)	
			data x=GetArg();		
			data y=GetArg();		
			Poke(x,y,m_WRegister);
		} break;
		
		case 0x03: 		
		{
			// movw x y (mem location to w)
			data x=GetArg();		
			data y=GetArg();		
			m_WRegister=Peek(x,y);
		} break;
		case 0x04: 		
		{
			// movif x y x y (indirect from w to address)
			data x1 = GetArg();
			data y1 = GetArg();
			data x2 = GetArg();
			data y2 = GetArg();
			data x=Peek(x1,y1);
			data y=Peek(x2,y2);			
			Poke(x,y,m_WRegister);
		} break;
		case 0x05: 		
		{
			// moviw x y x y (indirect from address to w)
			data x1 = GetArg();
			data y1 = GetArg();
			data x2 = GetArg();
			data y2 = GetArg();
			data x=Peek(x1,y1);
			data y=Peek(x2,y2);
			m_WRegister = Peek(x,y);
		} break;
		case 0x06: 		
		{
			// and x y
			data x=GetArg();		
			data y=GetArg();		
			m_WRegister &= Peek(x,y);
		} break;
		case 0x07: 		
		{
			// or x y
			data x=GetArg();		
			data y=GetArg();		
			m_WRegister |= Peek(x,y);
		} break;
		case 0x08: 		
		{
			// not x y
			data x=GetArg();		
			data y=GetArg();		
			m_WRegister =~ Peek(x,y);
		} break;
		case 0x09: 		
		{
			// xor x y
			data x=GetArg();		
			data y=GetArg();		
			m_WRegister ^= Peek(x,y);
		} break;
		case 0x0a: 		
		{
			// incw
			m_WRegister++;
		} break;
		case 0x0b: 		
		{
			// decw
			m_WRegister--;
		} break;
		case 0x0c: 		
		{
			// add x y
			data x=GetArg();	
			data y=GetArg();	
			m_WRegister += Peek(x,y);
		} break;
		case 0x0d: 		
		{
			// rotcw
			RotCW();
		} break;
		case 0x0e: 		
		{
			// rotccw
			RotCCW();
		} break;
		case 0x0f: 		
		{
			// jmp x y
			data x=GetArg();
			data y=GetArg();
			m_PCX=x;
			m_PCY=y;
		} break;
		case 0x10: 		
		{
			// rotcwz
			if (m_WRegister==0)
			{
				RotCW();
			}
		} break;
		case 0x11: 		
		{
			// rotccwz
			if (m_WRegister==0)
			{
				RotCCW();
			}
		} break;
		case 0x12: 		
		{
			// jmpz x y
			data x=GetArg();	
			data y=GetArg();
			if (m_WRegister==0)
			{
				m_PCX=x;
				m_PCY=y;
			}
		} break;
		case 0x13: 		
		{
			// threadspawncw 
			desc=SPAWN_CW;
			incpc=false;
		} break;
		case 0x14: 		
		{
			// threadspawnccw
			desc=SPAWN_CCW;
			incpc=false;
		} break;
		case 0x15: 		
		{
			// threadkill
			m_DelMe=true;
		} break;
		case 0x16: 		
		{
			// skipz
			if (m_WRegister==0)
			{
				IncPC();
			}
		} break;
		case 0x17: 		
		{
			// org
			unsigned int x=m_OriginX;
			unsigned int y=m_OriginY;
			GetWorldPos(x,y);
			m_OriginX=x;
			m_OriginY=y;
			m_PCX=0;
			m_PCY=0;
		} break;
		default: break;
	}
	if (incpc) IncPC();
}

//-------------------------------------------------------------------

char *PNGLoad(const string &Filename, int &w, int &h)
{
	char *ImageData = NULL;
	FILE *fp=fopen(Filename.c_str(),"rb");
	
	if (!fp) 
	{
		cerr<<"Couldn't open image ["<<Filename<<"]"<<endl;
	}
	else
	{
		png_structp png_ptr;
		png_infop info_ptr;
		
		png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
   	 	info_ptr = png_create_info_struct(png_ptr);
   		png_init_io(png_ptr, fp);
		png_read_png(png_ptr, info_ptr, PNG_TRANSFORM_IDENTITY, NULL);		
		fclose(fp);

		if (png_ptr->color_type!=PNG_COLOR_TYPE_GRAY)
		{
			cerr<<"quagmire thinks colour is frivolous"<<endl;
        }
		else
		{
			ImageData = new char[png_ptr->rowbytes*png_ptr->height];
			int p=0;
			for (unsigned int row = png_ptr->height-1; row>0 ; row--)
			{
				for (unsigned int i=0; i<png_ptr->rowbytes; i++)
				{
					ImageData[p]=(char)info_ptr->row_pointers[row][i];
					p++;
				}
			}

			w=png_ptr->width;
			h=png_ptr->height;
		}
		
		png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);	
	}
		
	return ImageData;
}

//-------------------------------------------------------------------

Quagmire *quag;
bool fast=false;
bool play=false;
bool step=false;
float zoom=1;

void DisplayCallback()
{
	if (play || step) 
	{
		quag->Process();
		step=false;
	}
    quag->Render(zoom);
	glutSwapBuffers();
	
	if (fast) glutPostRedisplay();	
}

void IdleCallback(int v)
{
	glutTimerFunc(100,IdleCallback,0);
	glutPostRedisplay();	
}

void KeyboardCallback(unsigned char key,int x, int y)
{
	if (key=='t') quag->RenderThreads(true);
	if (key=='c') quag->RenderThreads(false);
	if (key=='f') fast=!fast;
	if (key=='s') play=!play;
	if (key=='i') step=true;
}

int main(int argc, char *argv[])
{	
	if (argc!=3)
	{
		cerr<<"quagmire image.png zoom"<<endl;
		return -1;
	}
	
	int x=0,y=0;
	char *mem=PNGLoad(argv[1],x,y);	
	zoom = atof(argv[2]);
	if (mem==NULL) return -1;
	quag = new Quagmire((data*)mem,x,y);
	
	glutInitWindowSize((int)(x*zoom),(int)(y*zoom)) ;
  	glutInit(&argc,argv);
	glutInitDisplayMode(GLUT_DOUBLE|GLUT_RGBA|GLUT_DEPTH);
  	glutCreateWindow("quagmire");
	glutDisplayFunc(DisplayCallback);
	glutTimerFunc(1,IdleCallback,0);
	glutKeyboardFunc(KeyboardCallback);

	Matrix m;
	quag->SpawnThread(0,0,m);
	
	glutMainLoop();
    
	return 0;
}


